IMPLEMENTATION MODULE UserOps;
(*
 *  REPERTOIRE
 *  Release 1.5
 *  By Charles Bradford and Cole Brecheen
 *  (c) Copyright 1985, 1986, 1987, 1988 PMI
 *  Portland, Oregon
 *  All rights reserved
 *  (503) 777-8844
 *
 *)

IMPORT BigSets;
IMPORT ByteFiddler;
IMPORT ControlUtils;
IMPORT Editor;
IMPORT ErrorManager;
IMPORT GenLists;
IMPORT Hash;
IMPORT KbdInput;
IMPORT Key;
IMPORT ListUtils;
IMPORT M2Strings;
IMPORT MsgBox2;
IMPORT Numbers;
IMPORT NumTypes;
IMPORT PosUtils;
IMPORT ScrnTypes;
IMPORT ScrnUtl1;
IMPORT ScrnUtl2;
IMPORT StrConv;
IMPORT StrEdit;
IMPORT StrInput;
IMPORT Timing;
IMPORT VEditor;
IMPORT VStorage;
IMPORT VWindows;

  (* Temporary dependencies. *)
IMPORT WindowPrims;

IMPORT SYSTEM;

  PROCEDURE ShowMemory();
    VAR
      dumc: CARDINAL;
      tmpstr: ARRAY [0..31] OF CHAR;
      msgstr: ARRAY [0..127] OF CHAR;
    BEGIN
      IF VStorage.UsingEms() THEN
        msgstr := 'Expanded memory now available (in bytes):  ';
        StrConv.LongIntegerToStr( VStorage.AvailMem(), 0, tmpstr );
        StrEdit.Append( msgstr, tmpstr );
      ELSE
        msgstr := 'Expanded memory not in use.';
      END;
      tmpstr := 'Memory Status';
      dumc := VWindows.MessageBox( VWindows.CurrentWindow,
          SYSTEM.ADR(msgstr), SYSTEM.ADR(tmpstr), 0 );
    END ShowMemory;


CONST
  OldFunction = "Non-numeric frame name used with old function";

  PROCEDURE BetweenLongs( TheTxt: ARRAY OF CHAR; LoLong, HiLong: LONGINT) :
      BOOLEAN;
    VAR
      lvalue: LONGINT;
    BEGIN
      IF NOT StrConv.StrToLongInteger( TheTxt, 0, lvalue) THEN
        (*StrToLongInteger skips over leading non-numeric
          characters.*)
        RETURN FALSE;
      END;
      RETURN (lvalue >= LoLong) AND (lvalue <= HiLong);
    END BetweenLongs;


  PROCEDURE BetweenReals( TheTxt: ARRAY OF CHAR; LoReal, HiReal:
      NumTypes.Real8) : BOOLEAN;
    VAR
      rvalue: NumTypes.Real8;
    BEGIN
      IF NOT StrConv.StrToReal( TheTxt, 0, rvalue) THEN
        (*StrToReal skips over leading non-numeric
          characters.*)
        RETURN FALSE;
      END;
      RETURN (rvalue >= LoReal) AND (rvalue <= HiReal)
    END BetweenReals;


  PROCEDURE DefaultKeyHandler( FrameRec : ScrnTypes.DisplayFrame;
      VAR LastKey: CARDINAL; VAR NextFrame: ScrnTypes.AFrameName
      );
    VAR
      keyhit: CARDINAL;
      FieldRec: ScrnTypes.InputFieldRecord;
      StartFrame: ScrnTypes.AFrameName;
      BlankFrame : BOOLEAN;
      TmpScreen: WindowPrims.ScreenType;
    BEGIN
      StrEdit.AssignStr( ScrnTypes.ContinueInput, NextFrame );
      BlankFrame := ScrnUtl1.FieldListTotal( FrameRec ) = 0;
      IF NOT BlankFrame THEN
        ScrnUtl1.GetFieldRec( FrameRec, FrameRec^.CurrentField, FieldRec );
      END;
      IF LastKey = HelpKey THEN
        WindowPrims.ScreenSave( TmpScreen, 1, 1,
            VWindows.EndCol(VWindows.CurrentWindow),
            VWindows.EndRow(VWindows.CurrentWindow) );
          (* This screen save/restore is temporary--will be
             removed in the next release.  We recommend that you
             try to make your code work without it.*)
        IF (NOT BlankFrame) AND
            (M2Strings.Length(FieldRec.HelpFrame) > 0) THEN
          StartFrame := FieldRec.HelpFrame;
          ControlUtils.FrameSequence( FrameRec, StartFrame );
        ELSIF M2Strings.Length(FrameRec^.HelpFrame) > 0 THEN
          StartFrame := FrameRec^.HelpFrame;
          ControlUtils.FrameSequence( FrameRec, StartFrame );
        ELSIF NOT ListUtils.IsBlankList(FrameRec^.HelpList) THEN
          MsgBox2.LstMsgBox( FrameRec, VWindows.NW, FrameRec^.HelpList );
        END;
        WindowPrims.ScreenRestore( TmpScreen, 1, 1 );
        LastKey := 0;
          (*We don't want ControlFrame to do anything else with
            this keystroke.  We just want it to keep going.*)
      ELSIF LastKey = TimeKey THEN
(*
        Timing.ShowUsage('this program');
*)
        ShowMemory();
        LastKey := 0;
      ELSIF LastKey = BackupKey THEN
        M2Strings.Assign( FrameRec^.parent, NextFrame );
      ELSIF LastKey = SubmitKey THEN
        IF NOT BlankFrame THEN
          IF FieldRec.typ = ScrnTypes.GotoCode THEN
            M2Strings.Assign( FieldRec.ReturnVal, NextFrame );
          ELSE
            M2Strings.Assign( FrameRec^.normlnext, NextFrame );
          END;
        ELSE
          M2Strings.Assign( FrameRec^.normlnext, NextFrame );
        END;
      ELSIF LastKey = ExitKey THEN
        WindowPrims.MsgBox( VWindows.SW, EndPrompt,
            KbdInput.AnyKeyNum, keyhit);
        IF keyhit=ExitKey THEN
(*
          HALT;
*)
          ErrorManager.StraightToDOS('');
        END;
        LastKey := 0;
      END;
    END DefaultKeyHandler;


  PROCEDURE DefaultFieldHandler( FrameRec :
      ScrnTypes.DisplayFrame; col1, row1: CARDINAL; VAR LastKey :
      CARDINAL; VAR NextFrame: ScrnTypes.AFrameName );
    VAR
      ChangeCheck1, ChangeCheck2, ChangeCheck3, Check1, Check2,
          Check3, keyhit, CursorPos, width, spot: CARDINAL;
      done, InDecimals: BOOLEAN;
      LocalNextFrame: ScrnTypes.AFrameName;
      TmpLegalKeys, InterestingKeys: KbdInput.KeyNumSet;
      FieldRec : ScrnTypes.InputFieldRecord;
      ImageRec : ScrnTypes.ImageElement;
      TheList : GenLists.GenList;
      EdRec : VEditor.AnEdControlRec;

    BEGIN
      ScrnUtl1.GetFieldRec( FrameRec, FrameRec^.CurrentField, FieldRec );
      ScrnUtl1.GetFieldImageRec( FrameRec, FrameRec^.CurrentField, ImageRec );
      Hash.Compute( ImageRec.text, ChangeCheck1, ChangeCheck2, ChangeCheck3 );
      width := M2Strings.Length(ImageRec.text);
      InDecimals := FALSE;
      CursorPos := 0;
      BigSets.SetUnion( FieldExitKeySet, FunctKeySet, InterestingKeys );
      ScrnUtl1.SetPointerBarColors( FrameRec );

      IF FieldRec.typ = ScrnTypes.StringCode THEN
        TmpLegalKeys := StringKeySet;
          (* If you want to change which keys are legal in
             particular kinds of string fields, insert code here
             that looks for the name of the field you're
             interested in changing, and modify TmpLegalKeys with
             routines from BigSets if you find it. *)
        REPEAT
          ScrnUtl1.SetPointerBarColors( FrameRec );
          StrInput.ReadWithEdits( FrameRec^.WindowHandle,
              ImageRec.text, col1, row1, width, CursorPos,
              InsertMode, TRUE, LastKey, TmpLegalKeys,
              InterestingKeys );

            (* This is where you insert field validation code.
               It's also where you put code to update a frame
               dynamically as the user completes certain fields--
               use ChangeField, RedrawField, or RedrawArea to do
               the latter.*)

          Hash.Compute( ImageRec.text, Check1, Check2, Check3 );
          IF (ChangeCheck1 # Check1) OR (ChangeCheck2 # Check2)
              OR (ChangeCheck3 # Check3) THEN
            FieldRec.ChangeMade := TRUE;
          END;
          ScrnUtl1.PutFieldRec( FieldRec, FrameRec,
              FrameRec^.CurrentField );
          ScrnUtl1.PutFieldImageRec( ImageRec, FrameRec,
              FrameRec^.CurrentField );
          TheKeyHandler( FrameRec, LastKey, LocalNextFrame );
        UNTIL BigSets.InSet(FieldExitKeySet, LastKey);

      ELSIF FieldRec.typ = ScrnTypes.IntCode THEN
        done := FALSE;
        REPEAT
          ScrnUtl1.SetPointerBarColors( FrameRec );
          StrInput.ReadNumeric( FrameRec^.WindowHandle,
              ImageRec.text, col1, row1, width, 0, FALSE,
              InDecimals, LastKey, InterestingKeys );
            (* The user can't tell that we're exiting and re-
               entering ReadNumeric every time she presses one of
               the InterestingKeys. *)
          Hash.Compute( ImageRec.text, Check1, Check2, Check3 );
          IF (ChangeCheck1 # Check1) OR (ChangeCheck2 # Check2)
              OR (ChangeCheck3 # Check3) THEN
            FieldRec.ChangeMade := TRUE;
          END;
          ScrnUtl1.PutFieldRec( FieldRec, FrameRec,
              FrameRec^.CurrentField );
          ScrnUtl1.PutFieldImageRec( ImageRec, FrameRec,
              FrameRec^.CurrentField );
          TheKeyHandler( FrameRec, LastKey, LocalNextFrame );
          IF BigSets.InSet(FieldExitKeySet, LastKey) THEN
            IF (LastKey = ExitKey) OR (LastKey = BackupKey) THEN
              done := TRUE;
            ELSE
              done := PosUtils.IsBlank(ImageRec.text) OR
                 BetweenLongs( ImageRec.text, FieldRec.iMin,
                 FieldRec.iMax );
              IF NOT done THEN
                MsgBox2.StrMsgBox( FrameRec, VWindows.NE,
                    "Entry out of range.  Press <Enter> to continue." );
              END;
            END;
          END;
        UNTIL done;

      ELSIF FieldRec.typ = ScrnTypes.RealCode THEN
        done := FALSE;
        REPEAT
          ScrnUtl1.SetPointerBarColors( FrameRec );
          StrInput.ReadNumeric( FrameRec^.WindowHandle,
              ImageRec.text, col1, row1, width, 0, TRUE,
              InDecimals, LastKey, InterestingKeys );
          Hash.Compute( ImageRec.text, Check1, Check2, Check3 );
          IF (ChangeCheck1 # Check1) OR (ChangeCheck2 # Check2)
              OR (ChangeCheck3 # Check3) THEN
            FieldRec.ChangeMade := TRUE;
          END;
          ScrnUtl1.PutFieldRec( FieldRec, FrameRec,
              FrameRec^.CurrentField );
          ScrnUtl1.PutFieldImageRec( ImageRec, FrameRec,
              FrameRec^.CurrentField );
          TheKeyHandler( FrameRec, LastKey, LocalNextFrame );
          IF BigSets.InSet(FieldExitKeySet, LastKey) THEN
            IF (LastKey = ExitKey) OR (LastKey = BackupKey) THEN
              done := TRUE;
            ELSE
              done := PosUtils.IsBlank(ImageRec.text) OR
                 BetweenReals( ImageRec.text, FieldRec.rMin,
                 FieldRec.rMax );
              (* Note that we don't check the fractional part of
                 the number because we want to avoid making the
                 standard version of Repertoire depend on
                 Reals. You can add this form of checking by
                 putting real-number boundary fields in the
                 DisplayFrame record and modifying ScreenDisplay
                 to set them if the min and max numbers contain
                 decimal points.*)
              IF NOT done THEN
                MsgBox2.StrMsgBox( FrameRec, VWindows.NE,
                    "Entry out of range.  Press <Enter> to continue." );
              END;
            END;
          END;
        UNTIL done;

      ELSIF FieldRec.typ = ScrnTypes.EditorCode THEN
        ScrnUtl1.GetEdRec( FrameRec, FrameRec^.CurrentField, TheList, EdRec );
        EdRec.overwriting := NOT InsertMode;
        VEditor.TheEditor( TheList, EdRec );
        ScrnUtl1.PutEdRec( FrameRec, FrameRec^.CurrentField, TheList, EdRec );
        M2Strings.Assign( EdRec.NextFrame, NextFrame );
        InsertMode := NOT EdRec.overwriting;
        LastKey := EdRec.EdLastKey;
        RETURN;

      ELSE
        ErrorManager.WARN( 'Unexpected field type entered' );
      END;
      M2Strings.Assign( LocalNextFrame, NextFrame );
    END DefaultFieldHandler;


  PROCEDURE DoFunctions( FrameRec : ScrnTypes.DisplayFrame; VAR LastKey:
      CARDINAL ) : INTEGER;
    VAR
      NextFrame: ScrnTypes.AFrameName;
      tmpint: INTEGER;
    BEGIN
      NextFrame := '';
      DefaultKeyHandler( FrameRec, LastKey, NextFrame );
      IF PosUtils.Equal( NextFrame, ScrnTypes.ContinueInput ) THEN
        tmpint := ContinueCode;
      ELSIF NOT StrConv.StrToInteger( NextFrame, 0, tmpint ) THEN
        ErrorManager.WARN( OldFunction );
      END;
      RETURN tmpint;
    END DoFunctions;


  PROCEDURE DoFieldInput( FrameRec : ScrnTypes.DisplayFrame; col1, row1:
      CARDINAL; VAR LastKey : CARDINAL ) : INTEGER;
    VAR
      NextFrame: ScrnTypes.AFrameName;
      tmpint: INTEGER;
    BEGIN
      NextFrame := '';
      DefaultFieldHandler( FrameRec, col1, row1,
          LastKey, NextFrame );
      IF PosUtils.Equal( NextFrame, ScrnTypes.ContinueInput ) THEN
        tmpint := ContinueCode;
      ELSIF NOT StrConv.StrToInteger( NextFrame, 0, tmpint ) THEN
        ErrorManager.WARN( OldFunction );
      END;
      RETURN tmpint;
    END DoFieldInput;


  PROCEDURE OldKeyHandler( FrameRec : ScrnTypes.DisplayFrame; VAR
      LastKey: CARDINAL; VAR NextFrame: ScrnTypes.AFrameName );
    (*If you want to continue using procedures that you assigned
      to the Release 1.4 version of TheFunctProc, assign this
      procedure to TheKeyHandler.*)
    VAR
      tmpint: ScrnTypes.FrameKey;
    BEGIN
      tmpint := TheFunctProc( FrameRec, LastKey );
      IF tmpint = ContinueCode THEN
        StrEdit.AssignStr( ScrnTypes.ContinueInput, NextFrame );
      ELSE
        StrConv.IntegerToStr( tmpint, 0, NextFrame );
      END;
    END OldKeyHandler;

  PROCEDURE OldFieldHandler( FrameRec : ScrnTypes.DisplayFrame;
      col1, row1: CARDINAL; VAR LastKey : CARDINAL; VAR
      NextFrame: ScrnTypes.AFrameName );
    (*If you want to continue using procedures that you assigned
      to the Release 1.4 version of TheFieldProc, assign this
      procedure to TheFieldHandler.*)
    VAR
      tmpint: ScrnTypes.FrameKey;
    BEGIN
      tmpint := TheFieldProc( FrameRec, col1, row1, LastKey );
      IF tmpint = ContinueCode THEN
        StrEdit.AssignStr( ScrnTypes.ContinueInput, NextFrame );
      ELSE
        StrConv.IntegerToStr( tmpint, 0, NextFrame );
      END;
    END OldFieldHandler;

  BEGIN
    VEditor.TheEditor := Editor.DefaultEditor;
    Exiting := FALSE;
    InsertMode := FALSE;
    StrEdit.AssignStr(
        'Alt-D deletes lines; Alt-U undeletes; Shift-F7 toggles wrapping.',
        EdPrompt );
    StrEdit.AssignStr( 'Enter selects the highlighted choice.  ',JumpPrompt);
    StrConv.AppendDecoded( JumpPrompt, '#',
        'Use #024 or #025 to move to another choice.');
    StrEdit.SetLength(ChoicePrompt, 0);
    StrConv.AppendDecoded(ChoicePrompt, '#',
        'Use #017#196 or #196#016 to change the highlighted choice, or');
    StrConv.AppendDecoded(ChoicePrompt, '#',
        ' #024#025 to move to another field.');
    StrEdit.AssignStr('Fill in the blank and press',StringPrompt);
    StrConv.AppendDecoded(StringPrompt, '#',
        ' #024 or #025 to change fields, or F10 if finished.');
    StrEdit.AssignStr( 'Enter a number between MIN and MAX.',IntegerPrompt);
      (* MAX and MIN are replaced with the true values at
         runtime. *)
    M2Strings.Assign( IntegerPrompt, RealPrompt);
    M2Strings.Concat( 'Press the F2 key again to exit,',
        ' or press any other key to resume.', EndPrompt);
      (* the prompts that are printed for each type of input field *)

    DoPrompting := TRUE;
      (* whether or not to use prompts *)
    PromptRow := VWindows.EndRow(VWindows.CurrentWindow) - 1;
    PromptCol := 2;
      (* the screen row and column to show prompts if they are used *)
    PromptLength := VWindows.EndCol(VWindows.CurrentWindow) - 2;
      (* the maximum length of prompts used.  *)

    TheKeyHandler := DefaultKeyHandler;
    TheFieldHandler := DefaultFieldHandler;
      (* The routine that processes function keys in
         ControlFrame; see OldKeyHandler and OldFieldHandler,
         above, if you are having problems with upgrading 1.4
         code. *)

    TheFieldProc := DoFieldInput;
    TheFunctProc := DoFunctions;

    (* Now, define valid keys for function, screen actions,
       and string fields *)

    BigSets.InitSet( FunctKeySet );
    BigSets.AppendSet( FunctKeySet, "{27, 315, 316, 317, 324}" );
      (* extra keys we define as valid: ESC, F1, F2, F9, F10 *)

    BigSets.InitSet( ScreenKeySet );
      (* These are the keys that are legal when the pointer is
         not in an input field -- i.e., when it's in a choice
         field or goto field.  ScreenInput.ControlFrame adds to
         it the FunctKeySet and the selection characters for each
         frame. *)
    BigSets.AppendSet( ScreenKeySet,
        "{13, 9, 271, 328, 336, 329, 331, 333, 337, 374, 388}" );
      (* CR, TAB, BTB, UARW, DARW, PGUP, RARW, LARW, PGDN, CtrlPgDn, CtrlPgUp *)
    FieldExitKeySet := ScreenKeySet;
    BigSets.AppendSet( FieldExitKeySet, "{27, 317, 324}" );

    BigSets.InitSet( StringKeySet );
    BigSets.AppendSet( StringKeySet, "{8, ' '..'~', 128..175, 224..253}" );
    BigSets.AppendSet( StringKeySet,
        "{274, 288, 338, 339, 327, 335, 371, 372}" );
      (* regular ASCII characters, foreign language characters,
         math characters, backspace, AltE, AltD, Ins, Del,
         Home, End, CTL arrows *)

    HelpKey := Key.F1;
    ExitKey := Key.F2;
    TimeKey := Key.F3;
    SubmitKey := Key.F10;
    BackupKey := Key.Escape;
  END UserOps.
